from typing import Optional

import pandas as pd

from extralo.source import Source


class DeltaLakeSource(Source):
    """A source class for extracting data from Delta Lake tables.

    Args:
        table_uri (str): The URI of the Delta Lake table.
        partitions (Optional[list[tuple[str]]], optional): List of partition columns to filter the data.
            Defaults to None.
    """

    def __init__(
        self,
        table_uri: str,
        partitions: Optional[list[tuple[str]]] = None,
    ) -> None:
        try:
            import deltalake  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "DeltaLake is required to use DeltaLakeSource. Please install it with `pip install deltalake`."
            ) from err
        self._table_uri = table_uri
        self._partitions = partitions

    def extract(self) -> pd.DataFrame:
        """Extracts data from a Delta Lake table and returns it as a pandas DataFrame.

        Returns:
            DataFrame: The extracted data as a pandas DataFrame.
        """
        import deltalake as dl

        return dl.DeltaTable(self._table_uri).to_pandas(partitions=self._partitions)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(table_uri={self._table_uri})"



from .destinations import (
    CSVAppendDestination,
    CSVDestination,
    DeltaLakeDestination,
    SQLAppendDestination,
    SQLDestination,
    XLSXDestination,
)
from .etl import ETL
from .sources import CSVSource, DeltaLakeSource, SQLSource, XLSXSource

__all__ = [
    "ETL",
    "CSVSource",
    "SQLSource",
    "XLSXSource",
    "SQLDestination",
    "NullDestination",
    "CSVAppendDestination",
    "CSVDestination",
    "XLSXDestination",
    "SQLAppendDestination",
    "DeltaLakeDestination",
    "DeltaLakeSource",
]



import inspect
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional, _TypedDictMeta, get_type_hints  # type: ignore

from extralo.destination import Destination
from extralo.source import Source
from extralo.transformer import Transformer
from extralo.typing import DataFrame, DataFrameModel


class IncompatibleStepsError(Exception):
    """Exception raised when two steps in a process are incompatible.

    Attributes:
        step_base (str): The name of the base step.
        keys_base (set[str]): The keys associated with the base step.
        step (str): The name of the incompatible step.
        keys (set[str]): The keys associated with the incompatible step.
    """

    def __init__(self, step_base: str, keys_base: set[str], step: str, keys: set[str]) -> None:
        self.step_base = step_base
        self.step = step
        self.keys_base = keys_base
        self.keys = keys

    def __str__(self) -> str:
        return (
            f"Step '{self.step}' with keys {self.keys} is incompatible with step '{self.step_base}' "
            f"with keys {self.keys_base}"
        )


def _validate_steps(step1_keys: set[str], step1_name: str, step2_keys: set[str], step2_name: str) -> None:
    if step1_keys != step2_keys:
        raise IncompatibleStepsError(step1_name, step1_keys, step2_name, step2_keys)


def _validate_etl(sources_keys, before_schemas_keys, transform_method, after_schemas_keys, destinations_keys) -> None:
    if before_schemas_keys is not None:
        _validate_steps(sources_keys, "extract", before_schemas_keys, "before_schema")

    if transform_method is None:
        return

    trasnform_args = inspect.getargs(transform_method.__code__)

    if trasnform_args.varargs is not None:
        raise ValueError("Transformer transform method should not accept *args.")

    if trasnform_args.varkw is not None and len(trasnform_args.args) > 1:
        raise ValueError("Transformer transform method should only accept **kwargs or usally defined arguments.")

    if trasnform_args.varkw is None:
        _validate_steps(set(sources_keys), "extract", set(trasnform_args.args[1:]), "transform")

    transform_return_type_hint = get_type_hints(transform_method).get("return", None)
    no_return_type_hint = transform_return_type_hint is None
    return_type_hint_not_typed_dict = not isinstance(transform_return_type_hint, _TypedDictMeta)
    if no_return_type_hint or return_type_hint_not_typed_dict:
        logging.getLogger("etl").warning(
            "Transformer output type hints are not a TypedDict, validation will be done only at runtime."
        )
        return

    transform_output_dict = transform_return_type_hint.__annotations__
    if after_schemas_keys is not None:
        _validate_steps(set(transform_output_dict.keys()), "transform", after_schemas_keys, "after_schema")

    _validate_steps(set(transform_output_dict.keys()), "transform", destinations_keys, "load")


def _extract(source: Source) -> DataFrame:
    logger = logging.getLogger("etl")
    logger.info(f"Starting extraction for {source}")
    data = source.extract()
    logger.info(f"Extracted {len(data)} records from {source}")
    return data


def _validate(data: dict[str, DataFrame], schema: Optional[dict[str, DataFrameModel]]) -> dict[str, DataFrame]:
    if schema is None:
        logging.getLogger("etl").info("Skipping validation since no schema was provided.")
        return data

    return {name: schema.validate(data[name], lazy=True) for name, schema in schema.items()}


def _load(data: DataFrame, destination: Destination) -> None:
    logger = logging.getLogger("etl")
    logger.info(f"Starting load of {len(data)} records to {destination}")
    destination.load(data)
    logger.info(f"Loaded {len(data)} records to {destination}")


class ETL:
    """ETL - Extract, Load and Transform data from sources to destinations.

    The ETL class provide functionality around the extract, load and transform operations.
    The main functionalities are:

    - Ensure that the process will be executed in the correct order, and the data will only be loaded if satisfies the
        provided schema.
    - Allow the use of multiple sources.
    - Allow the use of multiple destinations for a single data.
    - Allow the use of different destinations for different data.
    - Provides configurable logging for each step of the process.
    - Run I/O operations in parallel, using threads.
    - Explicitly define where the data is comming from, what is happening with it and where it is going.

    The pipeline relies on dictionaries to define sources, validators and destinations. The keys of the dictionaries are
    used to match the data between the steps.

    In the earlier steps, it's possible to validate the match of the keys between the steps, to ensure that the data is
    being processed correctly.

    However, in the transform step, it's not possible to validated the keys, since the transformer can change the keys
    of the data. In this case, the validation will be done only at runtime.

    This can be resolved using return type annotations and the `TypedDict` from `typing` module. If the transformer
    return a `TypedDict`, the keys will be validated at initialization.

    Args:
        sources (dict[str, Source]): A dictionary with the sources to extract data from.
        destinations (dict[str, list[Destination]]): A dictionary with the destinations to load data to.
            Each value must be a list of destinations, and the data with that key will be loaded to all the
            destionations provided in the list.
        transformer (Transformer, optional): A transformer to transform the data. No transformation is done by default.
        before_schemas (Optional[dict[str, type[pa.DataFrameModel]]], optional): A dictionary with the schemas to
            validate the data before the transformation. Defaults to None.
        after_schemas (Optional[dict[str, type[pa.DataFrameModel]]], optional): A dictionary with the schemas to
            validate the data after the transformation. Defaults to None.
    """

    def __init__(
        self,
        sources: dict[str, Source],
        destinations: dict[str, list[Destination]],
        transformer: Optional[Transformer] = None,
        before_schemas: Optional[dict[str, DataFrameModel]] = None,
        after_schemas: Optional[dict[str, DataFrameModel]] = None,
    ) -> None:
        _validate_etl(
            set(sources.keys()),
            set(before_schemas.keys()) if before_schemas else None,
            transformer.transform if transformer else None,
            set(after_schemas.keys()) if after_schemas else None,
            set(destinations.keys()),
        )

        self._sources = sources
        self._destinations = destinations
        self._transformer = transformer
        self._before_schemas = before_schemas
        self._after_schemas = after_schemas

    def execute(self) -> None:
        """Execute the ETL process.

        Extract the data from the sources, validate it against the before schemas, transform it, validate it against
        the after schemas and load it to the destinations.

        Returns:
            This method does not return anything, and it's used for it's only side effect: load the data to the
                destinations.
        """
        data = self.extract()
        data = self.before_validate(data)
        data = self.transform(data)
        if self._after_schemas is not None:
            _validate_steps(set(data.keys()), "transform", set(self._after_schemas.keys()), "after_schema")
        data = self.after_validate(data)
        _validate_steps(set(data.keys()), "transform", set(self._destinations.keys()), "load")
        self.load(data)

    def extract(self) -> dict[str, DataFrame]:
        """Extract the data from the provided sources and load it into a dictionary with same keys as the sources.

        The extraction is done in parallel, using threads.
        This method use the `etl` logger to log the extraction process, which can be customized by the user.

        Returns:
            dict[str, DataFrame]: A dictionary with the data extracted from the sources.
        """
        with ThreadPoolExecutor(max_workers=5) as executor:
            extracted_data = executor.map(_extract, self._sources.values())
        names = self._sources.keys()
        return dict(zip(names, extracted_data))

    def before_validate(self, data: dict[str, DataFrame]) -> dict[str, DataFrame]:
        """Validate the extracted data against the provided schemas.

        If the before_schemas was not provided, no validation will be done and the data will be returned as is.

        Args:
            data (dict[str, DataFrame]): The data to be validated.

        Returns:
            dict[str, DataFrame]: A dictionary with the validated data (or the original data if no schema was provided).
                Uses the same keys as the input data.
        """
        return _validate(data, self._before_schemas)

    def transform(self, data: dict[str, DataFrame]) -> dict[str, DataFrame]:
        """Transform the data extracted from the source according to the `Transformer` class provided.

        This method use the `etl` logger to log the extraction process, which can be customized by the user.
        The data from the sources will be passes to the `transform` method of the `Transformer` as keyword arguments.

        Args:
            data (dict[str, DataFrame]): The data to be transformed.

        Returns:
            dict[str, DataFrame]: A dictionary with the transformed data. The keys could be different from the
                input data.
        """
        if self._transformer is None:
            logging.getLogger("etl").info("Skipping transform step since no Transformer was specified.")
            return data

        data = self._transformer.transform(**data)
        logging.getLogger("etl").info(f"Tranformed data with {self._transformer}")
        return data

    def after_validate(self, data: dict[str, DataFrame]) -> dict[str, DataFrame]:
        """Validate the transformed data against the provided schemas.

        If the after_schemas was not provided, no validation will be done and the data will be returned as is.

        Args:
            data (dict[str, DataFrame]): The data to be validated.

        Returns:
            dict[str, DataFrame]: A dictionary with the validated data (or the original data if no schema was provided).
                Uses the same keys as the input data.
        """
        return _validate(data, self._after_schemas)

    def load(self, data: dict[str, DataFrame]) -> None:
        """Load the data to the provided destinations.

        The data will be loaded in parallel, using threads.

        Args:
            data (dict[str, DataFrame]): The data to be loaded. The keys must match the keys of the destinations.
        """
        futures = []
        with ThreadPoolExecutor(max_workers=5) as executor:
            for name, destinations in self._destinations.items():
                data_to_load = data[name]
                for destination in destinations:
                    futures.append(executor.submit(_load, data_to_load, destination))

            for future in as_completed(futures):
                try:
                    future.result()
                except Exception as e:
                    raise Exception(f"Failed to load data: {e}") from e



from abc import ABC, abstractmethod

from extralo.typing import DataFrame


class Transformer(ABC):
    """Abstract base class for transformers.

    This class defines the interface for transformers in the extralo package.
    Subclasses should implement the `transform` method.
    """

    @abstractmethod
    def transform(self, **kwargs: DataFrame) -> dict[str, DataFrame]:
        """Transforms the input data using the specified transformations.

        Parameters:
            **kwargs (DataFrame): Input dataframes to be transformed.

        Returns:
            dict[str, DataFrame]: A dictionary containing the transformed dataframes.

        Raises:
            NotImplementedError: This method needs to be implemented in a subclass.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        return self.__repr__()



from abc import ABC, abstractmethod

from extralo.typing import DataFrame


class Destination(ABC):
    """Abstract base class for data sources."""

    @abstractmethod
    def load(self, data: DataFrame) -> None:
        """Load the given data into the destination.

        Args:
            data (DataFrame): The data to be loaded into the destination.

        Raises:
            NotImplementedError: This method is not implemented and should be overridden in a subclass.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        return self.__repr__()



from abc import ABC, abstractmethod

from extralo.typing import DataFrame


class Source(ABC):
    """Abstract base class for data sources."""

    @abstractmethod
    def extract(self) -> DataFrame:
        """Extracts data from a source and returns it as a DataFrame.

        Returns:
            DataFrame: The extracted data.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        return self.__repr__()



from typing import Any, Literal, Optional, Union

from extralo.destination import Destination
from extralo.typing import DataFrame


class DeltaLakeDestination(Destination):
    """A destination class for saving data to a Delta Lake table.

    Args:
        table_uri (str): The path to the Delta Lake table.
        **kwargs: Additional keyword arguments to be passed to the save function.
    """

    def __init__(
        self,
        table_uri: str,
        mode: Literal["error", "append", "overwrite", "ignore"],
        partition_by: Optional[Union[list[str], str]],
        **kwargs: Any,
    ) -> None:
        try:
            import deltalake  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "DeltaLake is required to use DeltaLakeDestination. Please install it with `pip install deltalake`."
            ) from err
        self._table_uri = table_uri
        self._mode = mode
        self._partition_by = partition_by
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(table={self._table_uri}, mode={self._mode})"

    def load(self, data: DataFrame) -> None:
        """Loads the given DataFrame into the Delta Lake table.

        Args:
            data (DataFrame): The DataFrame to be loaded.
        """
        import deltalake as dl

        dl.write_deltalake(
            table_or_uri=self._table_uri, data=data, mode=self._mode, partition_by=self._partition_by, **self._kwargs
        )



import os.path
from abc import ABC
from typing import Any

import pandas as pd

from extralo.destination import Destination


class FileDestination(Destination, ABC):
    """Represents a destination that writes data to a file.

    Args:
        file (str): The path to the file.
        **kwargs: Additional keyword arguments to be passed to the load function.
    """

    def __init__(self, file: str, **kwargs: Any) -> None:
        self._file = file
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(file={self._file})"


class CSVDestination(FileDestination):
    """A destination class for saving data to a CSV file."""

    def load(self, data: pd.DataFrame) -> None:
        """Save the given pandas DataFrame to a CSV file.

        If the file already exists, it will be overwritten.

        Args:
            data (DataFrame): The DataFrame to be saved.
        """
        data.to_csv(self._file, **self._kwargs)


class XLSXDestination(FileDestination):
    """A destination class for saving data to a XLSX file."""

    def load(self, data: pd.DataFrame) -> None:
        """Save the given pandas DataFrame to a XLSX file.

        If the file already exists, it will be overwritten.

        Args:
            data (DataFrame): The DataFrame to be saved.
        """
        data.to_excel(self._file, **self._kwargs)


class CSVAppendDestination(FileDestination):
    """A destination class for appending data to a CSV file."""

    def load(self, data: pd.DataFrame) -> None:
        """Append the given pandas DataFrame to a CSV file.

        If the file already exists, it will be appended, and will be assumed that the headers
        are in the same order.

        Args:
            data (DataFrame): The DataFrame to be saved.
        """
        if os.path.isfile(self._file):
            data.to_csv(self._file, mode="a", header=False, **self._kwargs)
            return
        data.to_csv(self._file, mode="w", header=True, **self._kwargs)



from typing import Any, Optional, Protocol

DataFrame = Any


class DataFrameModel(Protocol):  # noqa: D101
    def validate(  # noqa: D102
        self,
        check_obj: DataFrame,
        head: Optional[int] = None,
        tail: Optional[int] = None,
        sample: Optional[int] = None,
        random_state: Optional[int] = None,
        lazy: bool = False,
        inplace: bool = False,
    ) -> DataFrame: ...



# ETL:

# before validator deve validar todos as Keys q saíram do source e deve retornar
# trasnformer deve aceitar como input as mesmas Keys do sources.
# destinations deve aceitar como input as mesmas Keys exportadas do trasnformer

from typing import TypedDict
from unittest.mock import MagicMock, Mock

import pandas as pd
import pandera as pa
import pytest
from extralo.etl import ETL, IncompatibleStepsError


@pytest.fixture
def mock_data():
    return MagicMock(spec=pd.DataFrame)


@pytest.fixture
def mock_source(mock_data):
    class SourceStub:
        def extract(self):
            return mock_data

    return SourceStub()


@pytest.fixture
def mock_trans():
    class TransStub:
        def transform(self, source):
            return {"source_trans": source["source"]}

    return TransStub()


@pytest.fixture
def mock_dest():
    class DestStub:
        def load(self, data):
            return

    return DestStub()


@pytest.fixture
def mock_before_schema(mock_data):
    schema = MagicMock(spec=pa.DataFrameModel)
    schema.validate = Mock(return_value=mock_data)
    return schema


@pytest.fixture
def mock_after_schema(mock_data):
    schema = MagicMock(spec=pa.DataFrameModel)
    schema.validate = Mock(return_value=mock_data)
    return schema


def test_etl_delegates_to_components(mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema):
    etl = ETL(
        sources={"source": mock_source},
        transformer=mock_trans,
        destinations={"source_trans": [mock_dest]},
        before_schemas={"source": mock_before_schema},
        after_schemas={"source_trans": mock_after_schema},
    )

    etl.execute()

    assert True


def test_etl_works_without_transformer(mock_source, mock_dest, mock_before_schema):
    etl = ETL(
        sources={"source": mock_source},
        destinations={"source": [mock_dest]},
        before_schemas={"source": mock_before_schema},
    )

    etl.execute()

    assert True


def test_etl_works_without_before_schema(mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema):
    etl = ETL(
        sources={"source": mock_source},
        transformer=mock_trans,
        destinations={"source_trans": [mock_dest]},
        after_schemas={"source_trans": mock_after_schema},
    )

    etl.execute()

    assert True


def test_etl_works_without_after_schema(mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema):
    etl = ETL(
        sources={"source": mock_source},
        transformer=mock_trans,
        destinations={"source_trans": [mock_dest]},
        before_schemas={"source": mock_before_schema},
    )

    etl.execute()

    assert True


def test_etl_fails_when_source_and_before_validator_are_incompatible(
    mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema
):
    with pytest.raises(IncompatibleStepsError, match="extract"):
        ETL(
            sources={"source": mock_source},
            transformer=mock_trans,
            destinations={"source_trans": [mock_dest]},
            before_schemas={"not_source": mock_before_schema},
            after_schemas={"source_trans": mock_after_schema},
        )


def test_etl_fails_when_transformer_and_after_validator_are_incompatible(
    mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema
):
    etl = ETL(
        sources={"source": mock_source},
        transformer=mock_trans,
        destinations={"source_trans": [mock_dest]},
        before_schemas={"source": mock_before_schema},
        after_schemas={"not_source_trans": mock_after_schema},
    )
    with pytest.raises(IncompatibleStepsError, match="transform"):
        etl.execute()


def test_etl_fails_when_transformer_and_destination_are_incompatible(
    mock_source, mock_trans, mock_dest, mock_before_schema, mock_after_schema
):
    etl = ETL(
        sources={"source": mock_source},
        transformer=mock_trans,
        destinations={"not_source_trans": [mock_dest]},
        before_schemas={"source": mock_before_schema},
        after_schemas={"source_trans": mock_after_schema},
    )
    with pytest.raises(IncompatibleStepsError, match="load"):
        etl.execute()


def test_etl_fails_on_init_when_transformer_and_destination_are_incompatible_and_transformer_output_is_a_typed_dict(
    mock_source, mock_dest, mock_before_schema, mock_after_schema
):
    class TransformerOutputStub(TypedDict):
        source_trans: pd.DataFrame

    class TransformerStub:
        def transform(self, source) -> TransformerOutputStub:
            return source

    with pytest.raises(IncompatibleStepsError, match="load"):
        ETL(
            sources={"source": mock_source},
            transformer=TransformerStub(),
            destinations={"not_source_trans": [mock_dest]},
            before_schemas={"source": mock_before_schema},
            after_schemas={"source_trans": mock_after_schema},
        )


def test_etl_fails_on_init_when_transformer_and_after_validator_are_incompatible_and_transformer_output_is_a_typed_dict(
    mock_source, mock_dest, mock_before_schema, mock_after_schema
):
    class TransformerOutputStub(TypedDict):
        source_trans: pd.DataFrame

    class TransformerStub:
        def transform(self, source) -> TransformerOutputStub:
            return source

    with pytest.raises(IncompatibleStepsError, match="transform"):
        ETL(
            sources={"source": mock_source},
            transformer=TransformerStub(),
            destinations={"source_trans": [mock_dest]},
            before_schemas={"source": mock_before_schema},
            after_schemas={"not_source_trans": mock_after_schema},
        )


def test_etl_fails_when_source_and_transformer_are_incompatible(
    mock_source, mock_dest, mock_before_schema, mock_after_schema
):
    class TransformerStub:
        def transform(self, not_source):
            return not_source

    with pytest.raises(IncompatibleStepsError, match="transform"):
        ETL(
            sources={"source": mock_source},
            transformer=TransformerStub(),
            destinations={"source_trans": [mock_dest]},
            before_schemas={"source": mock_before_schema},
            after_schemas={"source_trans": mock_after_schema},
        )



from .delta_lake import DeltaLakeDestination
from .file import CSVAppendDestination, CSVDestination, XLSXDestination
from .sql import SQLAppendDestination, SQLDestination

__all__ = [
    "SQLDestination",
    "CSVDestination",
    "CSVAppendDestination",
    "XLSXDestination",
    "SQLAppendDestination",
    "DeltaLakeDestination",
]



from abc import ABC
from typing import Any

import pandas as pd

from extralo.source import Source
from extralo.typing import DataFrame


class FileSource(Source, ABC):
    """Represents a file source.

    Args:
        file (str): The path to the file.
        **kwargs: Additional keyword arguments to be passed to the read function.
    """

    def __init__(self, file: str, **kwargs: Any) -> None:
        self._file = file
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(file={self._file})"


class CSVSource(FileSource):
    """A class representing a CSV data source.

    This class inherits from the FileSource class and provides a method to extract data from a CSV file.
    """

    def __init__(self, file: str, **kwargs: Any) -> None:
        super().__init__(file, **kwargs)

    def extract(self) -> DataFrame:
        """Extracts data from a CSV file.

        Returns:
            DataFrame: The extracted data as a pandas DataFrame.
        """
        data: pd.DataFrame = pd.read_csv(self._file, **self._kwargs)
        return data


class XLSXSource(FileSource):
    """A class representing an XLSX data source.

    This class inherits from the FileSource class and provides a method to extract data from a XLSX file.
    """

    def __init__(self, file: str, **kwargs: Any) -> None:
        super().__init__(file, **kwargs)

    def extract(self) -> DataFrame:
        """Extracts data from an Excel file.

        Returns:
            DataFrame: The extracted data.
        """
        data: pd.DataFrame = pd.read_excel(self._file, **self._kwargs)
        return data



from typing import Any, Optional

import pandas as pd

from extralo.source import Source
from extralo.typing import DataFrame


class SQLSource(Source):
    """A class representing a SQL data source.

    Requires sqlparse and sqlalchemy to be installed. The SQL query is executed using the provided database engine.
    The query can have multiple statements, separated by semicolons. The last one must return some data.

    Args:
        engine (object): The database engine object.
        query (str): The SQL query to execute.
        params (dict, optional): The parameters to be passed to the SQL query. Defaults to None.
    """

    def __init__(self, engine: Any, query: str, params: Optional[dict[str, Any]] = None) -> None:
        try:
            import sqlalchemy as sa  # noqa: F401
            import sqlalchemy.exc as sa_exc  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "SQLAlchemy is required to use SQLSource. Please install it with `pip install sqlalchemy`."
            ) from err
        try:
            import sqlparse as sp  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "sqlparser is required to use SQLSource. Please install it with `pip install sqlparser`."
            ) from err
        self._engine = engine
        self._query = query
        self._params = params or {}

    def extract(self) -> DataFrame:
        """Extracts data from the database using the provided SQL query.

        Returns:
            DataFrame: The extracted data as a pandas DataFrame.
        """
        import sqlalchemy as sa
        import sqlalchemy.exc as sa_exc
        import sqlparse as sp

        query = sp.split(self._query)
        with self._engine.begin() as connection:
            for statement in query[:-1]:
                connection.execute(
                    sa.text(statement), parameters=self._params, execution_options={"no_parameters": True}
                )
            try:
                data = pd.read_sql(sa.text(query[-1]), connection, params=self._params)
            except sa_exc.ProgrammingError:
                data = pd.read_sql(sa.text(query[-1]), connection)

        return data

    def __repr__(self) -> str:
        return f"SQLSource(engine={self._engine})"



from typing import Any, Literal

import pandas as pd

from extralo.destination import Destination


class SQLDestination(Destination):
    """A class representing a SQL destination for loading data.

    Args:
        engine (sa.Engine): The SQLAlchemy engine to connect to the database.
        table (str): The name of the table to load the data into.
        schema (str): The name of the schema where the table resides.
        if_exists (str): The action to take if the table already exists.
    """

    def __init__(self, engine: Any, table: str, schema: str, if_exists: Literal["fail", "replace", "append"]) -> None:
        try:
            import sqlalchemy  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "SQLAlchemy is required to use SQLDestination. Please install it with `pip install sqlalchemy`."
            ) from err
        self._engine = engine
        self._table = table
        self._if_exists = if_exists
        self._schema = schema

    def load(self, data: pd.DataFrame) -> None:
        """Loads the given pandas DataFrame into an SQL table.

        Args:
            data (DataFrame): The pandas DataFrame to be loaded.
        """
        data.to_sql(name=self._table, schema=self._schema, con=self._engine, if_exists=self._if_exists, index=False)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(table={self._table}, schema={self._schema}, if_exists={self._if_exists})"


class SQLAppendDestination(SQLDestination):
    """A destination class for appending data to an SQL table, overriding by group.

    Args:
        engine (sa.Engine): The SQLAlchemy engine object.
        table (str): The name of the table.
        schema (str): The name of the schema.
        group_column (str): The name of the column used for grouping.
        group_value (Any): The value of the group column.
    """

    def __init__(self, engine: Any, table: str, schema: str, group_column: str, group_value: Any) -> None:
        try:
            import sqlalchemy  # noqa: F401
        except ImportError as err:
            raise ImportError(
                "SQLAlchemy is required to use SQLAppendDestination. Please install it with `pip install sqlalchemy`."
            ) from err

        super().__init__(engine, table, schema, "append")
        self._group_column = group_column
        self._group_value = group_value

    def load(self, data: pd.DataFrame) -> None:
        """Load data into the SQL table after deleting rows with a specific group value.

        Args:
            data (DataFrame): The data to be loaded into the table.

        Raises:
            KeyError: If the specified group column is not found in the table.
        """
        import sqlalchemy as sa

        insp = sa.inspect(self._engine)
        if insp.has_table(self._table, self._schema):
            metadata_obj = sa.MetaData()
            table = sa.Table(self._table, metadata_obj, autoload_with=self._engine, schema=self._schema)
            if self._group_column not in table.c:
                raise KeyError(f"Column '{self._group_column}' not found in table '{self._table}'")
            stmt = sa.delete(table).where(table.c[self._group_column] == self._group_value)
            with self._engine.begin() as conn:
                conn.execute(stmt)

        super().load(data)
